
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <string>
#include <queue>
#include <vector>

#include <SDL/SDL.h>
#include <SDL/SDL_sound.h>
#include <AL/al.h>

#include "Log.hh"
#include "Mippleton.hh"
#include "Sound.hh"

#define BUFFER_SIZE (64 * 1024)
//#define BUFFER_SIZE (5*2048)

namespace Mf {


struct Sound::Impl
{

	static ALenum getAudioFormat(const Sound_AudioInfo& audioInfo)
	{
		if (audioInfo.format == AUDIO_U8 || audioInfo.format == AUDIO_S8)
		{
			if (audioInfo.channels == 1) return AL_FORMAT_MONO8;
			else                         return AL_FORMAT_STEREO8;
		}
		else
		{
			if (audioInfo.channels == 1) return AL_FORMAT_MONO16;
			else                         return AL_FORMAT_STEREO16;
		}
	}
	
	class Buffer;
	typedef boost::shared_ptr<Buffer> BufferP;
	
	class Buffer : public Mippleton<Buffer>
	{
		Sound_Sample*			sound;
		std::vector<ALuint>		objects;

	public:

		Buffer(const std::string& name) :
			Mippleton<Buffer>(name),
			sound(0)
		{
			openFile();
		}

		~Buffer()
		{
			while (!objects.empty())
			{
				alDeleteBuffers(1, &objects.back());
				objects.pop_back();
			}

			if (sound) Sound_FreeSample(sound);
		}


		void openFile()
		{
			if (sound) Sound_FreeSample(sound);

			sound = Sound_NewSampleFromFile(Sound::getPath(getName()).c_str(),
					0, BUFFER_SIZE);

			if (!sound)
			{
				logWarning("error while loading sound %s: %s",
						getName().c_str(), Sound_GetError());
				throw Exception(Exception::FILE_NOT_FOUND);
			}

			logDebug("buffer size: %d", sound->buffer_size);
			logDebug("   channels: %d", sound->actual.channels);
			logDebug("     format: %d", sound->actual.format);
			logDebug("  frequency: %d", sound->actual.rate);
		}


		void loadAll(ALuint source)
		{
			if (!sound) openFile();
			if (!sound) return;

			unsigned decoded = Sound_DecodeAll(sound);
			if (decoded == 0)
			{
				logWarning("decoded no bytes from %s", getName().c_str());
				//throw Exception(Exception::FILE_NOT_FOUND);
				return;
			}

			ALuint obj;
			alGenBuffers(1, &obj);

			alBufferData(obj, getAudioFormat(sound->actual), sound->buffer,
					sound->buffer_size, sound->actual.rate);

			objects.push_back(obj);

			alSourcei(source, AL_BUFFER, obj);

			// don't need t his anymore
			Sound_FreeSample(sound);
			sound = 0;
		}


		void beginStream(ALuint source, int nBuffers = 4)
		{
			if (!sound) openFile();
			if (!sound) return;

			ALuint objs[nBuffers];
			alGenBuffers(nBuffers, objs);

			for (int i = 0; i < nBuffers; ++i)
			{
				objects.push_back(objs[i]);
				stream(objs[i]);
			}

			alSourceQueueBuffers(source, nBuffers, objs);
		}

		enum StreamStatus
		{
			STREAM_OK		= 0,
			STREAM_EOF		= 1,
			STREAM_WRONG	= 2
		};

		StreamStatus stream(ALuint buffer)
		{
			std::vector<ALuint>::iterator it =
				std::find(objects.begin(), objects.end(), buffer);

			// that buffer doesn't belong to us
			if (it == objects.end()) return STREAM_WRONG;

			unsigned bytes = Sound_Decode(sound);

			if (bytes == 0) return STREAM_EOF;

			alBufferData(buffer, getAudioFormat(sound->actual), sound->buffer,
					bytes, sound->actual.rate);

			return STREAM_OK;
		}

		inline void rewind()
		{
			if (!sound) openFile();
			else Sound_Rewind(sound);
		}


		// delete unused buffers, return true if all buffers deleted
		inline bool clear()
		{
			// clear any openal errors
			alGetError();

			while (!objects.empty())
			{
				ALuint buffer = objects.back();
				alDeleteBuffers(1, &buffer);

				// if an error occured, the buffer was not deleted because it's
				// still in use by some source
				if (alGetError() != AL_NO_ERROR) return false;

				objects.pop_back();
			}

			return true;
		}
	};


	Impl(const std::string& name) :
		buffer_(Buffer::getInstance(name)),
		playing_(false),
		looping_(false)
	{
		ALfloat zero[] = {0.0f, 0.0f, 0.0f};
		
		alGenSources(1, &source_);

		alSourcef(source_,  AL_PITCH, 1.0f);
		alSourcef(source_,  AL_GAIN, 1.0f);
		alSourcefv(source_, AL_POSITION, zero);
		alSourcefv(source_, AL_VELOCITY, zero);
	}

	~Impl()
	{
		alDeleteSources(1, &source_);
	}


	void play()
	{
		ALenum type;
		alGetSourcei(source_, AL_SOURCE_TYPE, &type);

		if (type != AL_STATIC)
		{
			buffer_->loadAll(source_);
		}

		alSourcei(source_, AL_LOOPING, looping_);
		alSourcePlay(source_);
		playing_ = true;
	}


	void stream()
	{
		ALenum type;
		alGetSourcei(source_, AL_SOURCE_TYPE, &type);

		alSourcei(source_, AL_BUFFER, AL_NONE);
		buffer_->rewind();
		buffer_->beginStream(source_);

		alSourcei(source_, AL_LOOPING, AL_FALSE);
		alSourcePlay(source_);
		playing_ = true;
	}

	inline void update()
	{
		ALint finished = 0;

		alGetSourcei(source_, AL_BUFFERS_PROCESSED, &finished);

		while (finished-- > 0)
		{
			ALuint buffer;

			alSourceUnqueueBuffers(source_, 1, &buffer);

			Buffer::StreamStatus status = buffer_->stream(buffer);

			if (status == Buffer::STREAM_OK)
			{
				alSourceQueueBuffers(source_, 1, &buffer);
			}
			else if (status == Buffer::STREAM_EOF)
			{
				if (!queue_.empty())
				{
					// begin the next buffer in the queue
					expired_.push_back(buffer_);
					buffer_ = queue_.front();
					queue_.pop();
					buffer_->beginStream(source_, 1);
				}
				else if (looping_)
				{
					// restart from the beginning
					buffer_->rewind();
					buffer_->stream(buffer);
					alSourceQueueBuffers(source_, 1, &buffer);
				}
			}
			else if (status == Buffer::STREAM_WRONG)
			{
				clear();
				buffer_->beginStream(source_, 1);
			}
		}

		ALenum state;
		alGetSourcei(source_, AL_SOURCE_STATE, &state);

		// restart playing if we're stopped but supposed to be playing... this
		// means we didn't queue enough and the audio skipped
		if (playing_ && state != AL_PLAYING)
		{
			alSourcePlay(source_);
		}
	}

	inline void clear()
	{
		// try to remove expired buffers
		std::vector<BufferP>::iterator it;
		for (it = expired_.end() - 1; it >= expired_.begin(); --it)
		{
			if ((*it)->clear()) expired_.erase(it);
		}
	}


	void stop()
	{
		alSourceStop(source_);
		playing_ = false;
	}

	inline void pause()
	{
		alSourcePause(source_);
		playing_ = false;
	}

	inline void resume()
	{
		alSourcePlay(source_);
		playing_ = true;
	}


	inline void setSample(const std::string& name)
	{
		bool playing = isPlaying();
		ALenum type;
		alGetSourcei(source_, AL_SOURCE_TYPE, &type);

		stop();

		//alSourcei(source_, AL_BUFFER, AL_NONE);
		buffer_ = Buffer::getInstance(name);

		if (type == AL_STREAMING)
		{
			if (playing) stream();
		}
		else
		{
			if (playing) play();
		}
	}

	inline void enqueue(const std::string& name)
	{
		BufferP buffer = Buffer::getInstance(name);
		queue_.push(buffer);
	}


	inline bool isPlaying() const
	{
		if (playing_) return true;

		ALenum state;
		alGetSourcei(source_, AL_SOURCE_STATE, &state);

		return state == AL_PLAYING;
	}


	inline void setLooping(bool looping)
	{
		looping_ = looping;

		ALenum type;
		alGetSourcei(source_, AL_SOURCE_TYPE, &type);

		if (type != AL_STREAMING)
		{
			alSourcei(source_, AL_LOOPING, looping_);
		}
	}


	ALuint					source_;
	BufferP					buffer_;

	bool					playing_;
	bool					looping_;

	std::queue<BufferP>		queue_;
	std::vector<BufferP>	expired_;
};

Sound::Sound(const std::string& name) :
	// pass through
	impl_(new Sound::Impl(name)) {}


void Sound::play()
{
	// pass through
	impl_->play();
}


void Sound::stream()
{
	// pass through
	impl_->stream();
}

void Sound::update(Scalar t, Scalar dt)
{
	// pass through
	impl_->update();
}


void Sound::stop()
{
	// pass through
	impl_->stop();
}

void Sound::pause()
{
	// pass through
	impl_->pause();
}

void Sound::resume()
{
	// pass through
	impl_->resume();
}

void Sound::toggle()
{
	if (impl_->playing_) pause();
	else resume();
}


void Sound::setSample(const std::string& name)
{
	// pass through
	impl_->setSample(name);
}

void Sound::enqueue(const std::string& name)
{
	// pass through
	impl_->enqueue(name);
}


bool Sound::isPlaying() const
{
	// pass through
	return impl_->isPlaying();
}

void Sound::setPosition(Vector3 position)
{
	float p[3] = {position[0], position[1], position[2]};
	alSourcefv(impl_->source_, AL_POSITION, p);
}

void Sound::setVelocity(Vector3 velocity)
{
	float v[3] = {velocity[0], velocity[1], velocity[2]};
	alSourcefv(impl_->source_, AL_VELOCITY, v);
}

void Sound::setGain(Scalar gain)
{
	alSourcef(impl_->source_, AL_GAIN, float(gain));
}

void Sound::setPitch(Scalar pitch)
{
	alSourcef(impl_->source_, AL_PITCH, float(pitch));
}

void Sound::setLooping(bool looping)
{
	// pass through
	impl_->setLooping(looping);
}


std::string Sound::getPath(const std::string& name)
{
	std::string path = Resource::getPath("sounds/" + name + ".ogg");
	return path;
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

